/* -*- mode: c; c-basic-offset: 4; -*- */
/* vi:set cindent ts=8 sts=4 sw=4: */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <ncurses.h>
#include <string.h>
#include <assert.h>
#include <limits.h>

#include "main.h"

#include "curs_color.h"
#include "curs_lib.h"
#include "curs_main.h"
#include "curs_menu.h"
#include "keymap.h"
#include "lib.h"
#include "misc.h"
#include "options.h"
#include "playlist.h"

static void menu_pad_string(char *str, size_t len);
static int enter_index_number(int *dst, char *prompt, int current,
			      int entries);

#define RETURN_IF_NO_ENTRIES \
if (!menu->entries) { msg("There are no entries in the menu."); return; }

void menu_browser_help_make_entry(char *str, size_t len, int index)
{
    int i = 0, j = 0;

    while (i <= index)
	if (keymap[j++].menus & MENU_BROWSER)
	    i++;
    j--;

    snprintf(str, len, "%s", keymap[j].name);
    /* two spaces between the maximum command name length + the
       terminating \0 */
    if (len > MAX_CMD_NAME_LEN) {
	char key_descr[SHORT_STRING];
	key_description(j, key_descr, sizeof(key_descr));
	menu_pad_string(str, MAX_CMD_NAME_LEN + 3);
	snprintf(str + strlen(str), len - strlen(str), "%s", key_descr);
    }
}

void menu_options_help_make_entry(char *str, size_t len, int index)
{
    int i = 0, j = 0;

    while (i <= index)
	if (keymap[j++].menus & MENU_OPTIONS)
	    i++;
    j--;

    snprintf(str, len, "%s", keymap[j].name);
    /* two spaces between the maximum command name length + the
       terminating \0 */
    if (len > MAX_CMD_NAME_LEN) {
	char key_descr[SHORT_STRING];
	key_description(j, key_descr, sizeof(key_descr));
	menu_pad_string(str, MAX_CMD_NAME_LEN + 3);
	snprintf(str + strlen(str), len - strlen(str), "%s", key_descr);
    }
}

void menu_main_help_make_entry(char *str, size_t len, int index)
{
    int i = 0, j = 0;

    while (i <= index)
	if (keymap[j++].menus & MENU_MAIN)
	    i++;
    j--;

    snprintf(str, len, "%s", keymap[j].name);
    /* two spaces between the maximum command name length + the
       terminating \0 */
    if (len > MAX_CMD_NAME_LEN) {
	char key_descr[SHORT_STRING];
	key_description(j, key_descr, sizeof(key_descr));
	menu_pad_string(str, MAX_CMD_NAME_LEN + 3);
	snprintf(str + strlen(str), len - strlen(str), "%s", key_descr);
    }
}

static void menu_pad_string(char *str, size_t len)
{
    int n = strlen(str);

    len--;			/* save room for the terminal \0 */
    if (len > COLS)
	len = COLS;
    /* we have to pad the string with blanks to the end of line */
    if (n < len) {
	while (n < len)
	    str[n++] = ' ';
	str[n] = 0;
    }
}

/*
 * Get and parse command from the user.
 * Set `cmd' to the identification number of the command received, and
 * `num_cmd_arg' to the numeric argument (0 if not given).
 * Do not return before the user types a valid command.
 */
void menu_get_command(const struct menu *menu, int *cmd, int *num_cmd_arg)
{
    int key;
    struct keymap_entry *p;

    /* One or more of the value `ERR' is sometimes pending in the input queue
       because we use halfdelay() in play_toggled_songs() */
    while ((key = getch()) == ERR)
	;
    (void) ungetch(key);

 start_over_again:

    /* Parse numeric argument */
    clear_accumulator();
    while (isdigit(key = getch()))
	add_to_accumulator(key);

    /* Two `ESC's in a row mean cancel the current command */
    if (key == ESC) {
	key = getch();
	if (key == ESC)
	    goto start_over_again;
	(void) ungetch(key);
	key = ESC;
    }

    /* Scan through the keymap, looking for matching entries */
    p = keymap;
    while (p->cmd_id != KEYMAP_TERMINATOR) {
	/* If the keymap entry is valid for the current menu */
	if (p->menus & menu->menu) {
	    if (p->firstchar == key) {
		if (p->nextchar == 0) {
		    *num_cmd_arg = accumulator_value();
		    *cmd = p->cmd_id;
		    return;
		} else {
		    break;
		}
	    }
	}
	p++;
    }
    /* If `key' is not equal to the first character of any keymap entries */
    if (p->cmd_id == KEYMAP_TERMINATOR) {
	msg("Unknown command.");
	refresh();
	goto start_over_again;
    }

    /* Two `ESC's in a row mean cancel the current command */
    key = getch();
    if (key == ESC) {
	key = getch();
	if (key == ESC)
	    goto start_over_again;
	(void) ungetch(key);
	key = ESC;
    }

    /* Scan through the keymap, looking for matching entries */
    p = keymap;
    while (p->cmd_id != KEYMAP_TERMINATOR) {
	if ((p->menus & menu->menu) && p->nextchar == key) {
	    *num_cmd_arg = accumulator_value();
	    *cmd = p->cmd_id;
	    return;
	}
	p++;
    }
    msg("Unknown command.");
    refresh();
    goto start_over_again;
}

void menu_redraw_index(struct menu *menu)
{
    char buf[STRING];
    int i;

    attrset(main_attr);

    for (i = menu->top; i < menu->top + menu->pagelen; i++) {
	if (i < menu->entries) {
	    menu->make_entry(buf, sizeof(buf), i);
	    menu_pad_string(buf, sizeof(buf));
	    move(i - menu->top + menu->offset, 0);
	    if (i == menu->current) {
		attrset(current_entry_attr);
		addnstr(buf, COLS);
		attrset(main_attr);
	    } else
		addnstr(buf, COLS);
	} else
	    CLEARLINE(i - menu->top + menu->offset);
    }
}

void menu_redraw_motion(struct menu *menu)
{
    char buf[STRING];

    /* erase the current indicator */
    menu->make_entry(buf, sizeof(buf), menu->oldcurrent);
    menu_pad_string(buf, sizeof(buf));
    move(menu->oldcurrent + menu->offset - menu->top, 0);
    attrset(main_attr);
    addnstr(buf, COLS);

    /* now draw the new one to reflect the change */
    menu->make_entry(buf, sizeof(buf), menu->current);
    menu_pad_string(buf, sizeof(buf));
    move(menu->current - menu->top + menu->offset, 0);
    attrset(current_entry_attr);
    addnstr(buf, COLS);
}

void menu_redraw_current(struct menu *menu)
{
    char buf[STRING];

    move(menu->current + menu->offset - menu->top, 0);
    menu->make_entry(buf, sizeof(buf), menu->current);
    menu_pad_string(buf, sizeof(buf));
    attrset(current_entry_attr);
    addnstr(buf, COLS);
}

/*
 * Redraw the status bar.
 */
void menu_redraw_status(const struct menu *menu, const char *playlist,
		 bool continuous, bool random)
{
    char status[STRING];

    if (menu->menu == MENU_MAIN) {
	snprintf(status, sizeof(status), "%s  c:%s r:%s  %s", PROGNAME,
		 continuous ? "on" : "off", random ? "on" : "off",
		 playlist ? playlist : "");
    } else if (menu->menu == MENU_OPTIONS) {
	snprintf(status, sizeof(status), "Options");
    } else if (menu->menu == MENU_HELP) {
	snprintf(status, sizeof(status), "Help");
    } else if (menu->menu == MENU_BROWSER) {
	snprintf(status, sizeof(status), "Browser  %s", cwd());
    } else
	assert(0);

    if (strlen(status) > COLS)
        strcpy(status + COLS - 3, "...");
    menu_pad_string(status, sizeof(status));
    attrset(status_attr);
    mvaddnstr(STATUS_LINE, 0, status, COLS);
}

/*
 * Redraw the command reference.
 */
void menu_redraw_cmd_ref(const struct menu *menu)
{
    char buf[STRING];

    if (menu->menu == MENU_MAIN)
	snprintf(buf, sizeof(buf), command_reference(MENU_MAIN));
    else if (menu->menu == MENU_HELP)
	snprintf(buf, sizeof(buf), command_reference(MENU_HELP));
    else if (menu->menu == MENU_BROWSER)
	snprintf(buf, sizeof(buf), command_reference(MENU_BROWSER));
    else if (menu->menu == MENU_OPTIONS)
	snprintf(buf, sizeof(buf), command_reference(MENU_OPTIONS));
    else
	assert(0);

    menu_pad_string(buf, sizeof(buf));
    attrset(cmd_ref_attr);
    mvaddnstr(CMD_REF_LINE, 0, buf, COLS);
}

void cmd_page_down(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == menu->entries - 1)
	return;

    if (menu->top + menu->pagelen < menu->entries) {
	menu->top += menu->pagelen;
	menu->current = menu->top;
    } else { /* We're on the last page */
	menu->current = menu->entries - 1;
    }
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

void cmd_page_up(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == 0)
	return;

    if (menu->top > 0) {
	menu->top -= menu->pagelen;
	if (menu->top < 0)
	    menu->top = 0;
	if (menu->current > menu->top + menu->pagelen - 1)
	    menu->current = menu->top + menu->pagelen - 1;
    } else { /* We're on the first page */
	menu->current = menu->top;
    }
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

/*
 * Very much the same as cmd_move_song_to_top_of_page().
 */
void cmd_top_of_page(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == menu->top)
	return;

    menu->current = menu->top;
    TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
}

/*
 * Very much the same as cmd_top_of_page().
 */
void cmd_move_song_to_top_of_page(struct menu *menu, bool *playlist_changed)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == menu->top)
	return;

    move_entry(menu->top, menu->current);
    menu->current = menu->top;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

/*
 * Very much the same as cmd_move_song_to_bottom_of_page().
 */
void cmd_bottom_of_page(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == menu->top + menu->pagelen - 1
	|| menu->current == menu->entries - 1)
	return;

    menu->current = menu->top + menu->pagelen - 1;
    if (menu->current > menu->entries - 1)
	menu->current = menu->entries - 1;
    TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
}

/*
 * Very much the same as cmd_bottom_of_page().
 */
void cmd_move_song_to_bottom_of_page(struct menu *menu, bool *playlist_changed)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == menu->top + menu->pagelen - 1
	|| menu->current == menu->entries - 1)
	return;

    if (menu->top + menu->pagelen - 1 > menu->entries - 1) {
	move_entry(menu->entries - 1, menu->current);
	menu->current = menu->entries - 1;
    } else {
	move_entry(menu->top + menu->pagelen - 1, menu->current);
	menu->current = menu->top + menu->pagelen - 1;
    }
    *playlist_changed = TRUE;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

/*
 * Very much the same as cmd_move_song_to_middle_of_page().
 */
void cmd_middle_of_page(struct menu *menu)
{
    int dst;

    RETURN_IF_NO_ENTRIES;

    /* If there's an even number of entries on the page, there are two
       entries in the middle of the page. In that case, we refer to the
       first of those two entries as the middle of the page */
    dst = (menu->top + menu->pagelen <= menu->entries - 1) ?
	menu->top + (menu->pagelen - 1) / 2 :
	menu->top + (menu->entries - 1 - menu->top) / 2;

    if (menu->current == dst)
	return;

    menu->current = dst;
    TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
}

/*
 * Very much the same as cmd_middle_of_page).
 */
void cmd_move_song_to_middle_of_page(struct menu *menu, bool *playlist_changed)
{
    int dst;

    RETURN_IF_NO_ENTRIES;

    /* If there's an even number of entries on the page, there are two
       entries in the middle of the page. In that case, we refer to the
       first of those two entries as the middle of the page */
    dst = (menu->top + menu->pagelen <= menu->entries - 1) ?
	menu->top + (menu->pagelen - 1) / 2 :
	menu->top + (menu->entries - 1 - menu->top) / 2;

    if (menu->current == dst)
	return;

    move_entry(dst, menu->current);
    *playlist_changed = TRUE;
    menu->current = dst;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

/*
 * Very much the same as cmd_move_song_first().
 */
void cmd_first_entry(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == 0)
	return;

    if (menu->top > 0)
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
    else
	TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);

    menu->top = menu->current = 0;
}

/*
 * Very much the same as cmd_first_entry().
 */
void cmd_move_song_first(struct menu *menu, bool *playlist_changed)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == 0)
	return;

    move_entry(0, menu->current);
    *playlist_changed = TRUE;
    menu->top = menu->current = 0;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

/*
 * Very much the same as cmd_move_song_last().
 */
void cmd_last_entry(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == menu->entries - 1)
	return;

    menu->current = menu->entries - 1;
    if (menu->current > menu->top + menu->pagelen - 1) {
	menu->top = menu->current - menu->pagelen + 1;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
    } else {
	TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
    }
}

/*
 * Very much the same as cmd_last_entry().
 */
void cmd_move_song_last(struct menu *menu, bool *playlist_changed)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == menu->entries - 1)
	return;

    move_entry(menu->entries - 1, menu->current);
    *playlist_changed = TRUE;
    menu->current = menu->entries - 1;
    if (menu->current > menu->top + menu->pagelen - 1)
	menu->top = menu->current - menu->pagelen + 1;

    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

void cmd_half_page_down(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == menu->entries - 1)
	return;

    if (menu->top + menu->pagelen >= menu->entries) {
	menu->current += menu->pagelen / 2;
	if (menu->current > menu->entries - 1)
	    menu->current = menu->entries - 1;
	TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
    } else if (menu->current + menu->pagelen / 2 > menu->entries - 1) {
	menu->current = menu->entries - 1;
	menu->top = menu->entries - menu->pagelen;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
    } else {
	menu->top += menu->pagelen / 2;
	if (menu->top + menu->pagelen > menu->entries)
	    menu->top = menu->entries - menu->pagelen;
	menu->current += menu->pagelen / 2;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
    }
}

void cmd_half_page_up(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == 0)
	return;

    menu->current -= menu->pagelen / 2;
    /* Justify if we added too much */
    if (menu->current < 0)
	menu->current = 0;

    /* If the first entry is not shown */
    if (menu->top > 0) {
	menu->top -= menu->pagelen / 2;
	/* Justify if we subtracted too much */
	if (menu->top < 0)
	    menu->top = 0;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
    } else
	TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
}

void cmd_scroll_down(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->top >= menu->entries - 1)
	return;

    menu->top++;
    if (menu->current < menu->top)
	menu->current++;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

void cmd_scroll_up(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->top <= 0)
	return;

    menu->top--;
    if (menu->current > menu->top + menu->pagelen - 1)
	menu->current--;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

/*
 * Very much the same as cmd_move_song_up().
 */
void cmd_previous_entry(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current <= 0)
	return;

    menu->current--;
    TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
}

/*
 * Very much the same as cmd_previous_entry().
 */
void cmd_move_song_up(struct menu *menu, bool *playlist_changed)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current <= 0)
	return;

    move_entry(menu->current - 1, menu->current);
    menu->current--;
    *playlist_changed = TRUE;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

/*
 * Very much the same as cmd_move_song_down().
 */
void cmd_next_entry(struct menu *menu)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == menu->entries - 1)
	return;

    menu->current++;
    TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
}

/*
 * Very much the same as cmd_next_entry().
 */
void cmd_move_song_down(struct menu *menu, bool *playlist_changed)
{
    RETURN_IF_NO_ENTRIES;

    if (menu->current == menu->entries - 1)
	return;

    move_entry(menu->current + 1, menu->current);
    menu->current++;
    *playlist_changed = TRUE;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

/*
 * Get input from the user, calculate a valid index number from it, and write
 * the number into `dst'.
 * Return 0 if an index number is written into `dst', or -1 if not.
 */
static int enter_index_number(int *dst, char *prompt, int current, int entries)
{
    char buf[SHORT_STRING];
    int i;
    bool leading_minus = FALSE;
    bool leading_plus = FALSE;

    if (enter_string(buf, sizeof(buf), EMPTY, NO_BROWSE, prompt) == NULL)
	return -1;
    /* If the user only hit linefeed */
    if (buf[0] == '\0')
	return -1;

    if (buf[0] == '-') {
	leading_minus = TRUE;
    } else if (buf[0] == '+') {
	leading_plus = TRUE;
    }
    if (leading_minus || leading_plus) {
	if (buf[1] == '\0') {
	    msg("Invalid number.");
	    return -1;
	}
    }

    if (leading_minus || leading_plus)
	i = 1;
    else
	i = 0;
    clear_accumulator();
    while (isdigit(buf[i])) {
	add_to_accumulator(buf[i]);
	i++;
    }
    if (buf[i] != '\0') {
	msg("Invalid number.");
	return -1;
    }

    /* If the index number given is not relative to the current entry. */
    if (!leading_minus && !leading_plus) {
	*dst = accumulator_value();
	(*dst)--; /* To the user, the indexing starts at 1, but internally, the
		     indexing starts at 0 */
	if (*dst > entries - 1)
	    *dst = entries - 1;
	else if (*dst < 0)
	    *dst = 0;
    } else if (leading_minus == TRUE) {
	*dst = current - accumulator_value();
	if (*dst < 0)
	    *dst = 0;
    } else { /* leading_plus == TRUE */
	*dst = current + accumulator_value();
	if (*dst > entries - 1)
	    *dst = entries - 1;
    }

    return 0;
}

/*
 * Very much the same as cmd_move_song_to().
 */
void cmd_go_to(struct menu *menu)
{
    int dst;

    if (enter_index_number(&dst, "Go to: ", menu->current,
			   menu->entries) == -1)
	return;
    if (menu->current == dst)
	return;
    
    if (dst > menu->top + menu->pagelen - 1) {
	menu->top = dst - menu->pagelen + 1;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
    } else if (dst < menu->top) {
	menu->top = dst;
	TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
    } else {
	TURN_ON_FLAG(menu->redraw, REDRAW_MOTION);
    }
    menu->current = dst;
}

/*
 * Very much the same as cmd_go_to().
 */
void cmd_move_song_to(struct menu *menu, bool *playlist_changed)
{
    int dst;

    if (enter_index_number(&dst, "Move the current song to: ", menu->current,
			   menu->entries) == -1)
	return;
    if (menu->current == dst)
	return;

    if (dst > menu->top + menu->pagelen - 1) {
	menu->top = dst - menu->pagelen + 1;
    } else if (dst < menu->top) {
	menu->top = dst;
    }
    move_entry(dst, menu->current);
    *playlist_changed = TRUE;
    TURN_ON_FLAG(menu->redraw, REDRAW_INDEX);
}

