/* -*- mode: c; c-basic-offset: 4; -*- */
/* vi:set cindent ts=8 sts=4 sw=4: */

#include <stdio.h>
#include <stdarg.h>
#include <ncurses.h>
#include <sys/stat.h>
#include <limits.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <pwd.h>
#include <assert.h>

#include "main.h"

#include "curs_lib.h"
#include "lib.h"
#include "misc.h"

/* Accumulator */
static int acc;

/*
 * Set the static accumulator variable to 0.
 */
void clear_accumulator(void)
{
    acc = 0;
}

/*
 * If `value' is a digit in the ASCII, add it to the static accumulator
 * variable, which will be set to `INT_MAX' on overflow.
 */
void add_to_accumulator(int value)
{
    static int cutoff = INT_MAX / 10;
    static int cutlim = INT_MAX % 10;

    if (isdigit(value)) {
	value -= '0';
	if (acc > cutoff || (acc == cutoff && value > cutlim)) {
	    acc = INT_MAX;
	} else {
	    acc *= 10;
	    acc += value;
	}
    }
}

/*
 * Return the value of the static accumulator variable.
 */
int accumulator_value(void)
{
return acc;
}

void *safe_calloc(size_t nmemb, size_t size)
{
    void *p;

    if (!nmemb || !size)
	return NULL;
    if ((p = calloc(nmemb, size)) == NULL)
	error_exit("Out of memory!");

    return p;
}

void *safe_malloc(size_t size)
{
    void *p;

    if (size == 0)
	return NULL;
    if ((p = malloc(size)) == NULL)
	error_exit("Out of memory!");

    return p;
}

void safe_realloc(void **p, size_t size)
{
    void *r = NULL;

    if (size == 0) {
	if (*p) {
	    free(*p);
	    *p = NULL;
	}
	return;
    }
    r = (void *) realloc(*p, size);
    if (!r)
	error_exit("Out of memory!");

    *p = r;
}

char *safe_strdup(const char *s)
{
    char *p;

    if (!s || !*s)
	return NULL;
    p = safe_malloc(strlen(s) + 1);
    strcpy(p, s);

    return p;
}

char *strstr_icase(char *haystack, char *needle)
{
    char *cp = haystack;

    while ((cp = strchr(cp, needle[0]))) {
	if (strncasecmp(cp, needle, strlen(needle)) == 0)
	    return cp;
	cp++;
    }

    cp = haystack;
    if (islower(needle[0])) {
	while ((cp = strchr(cp, toupper(needle[0])))) {
	      if (strncasecmp(cp, needle, strlen(needle)) == 0)
		return cp;
	    cp++;
	}
    } else if (isupper(needle[0])) {
	while ((cp = strchr(cp, tolower(needle[0])))) {
	if (strncasecmp(cp, needle, strlen(needle)) == 0)
		return cp;
	    cp++;
	}
    }
    return NULL;
}

/*
 * Remove leading and trailing whitespace from str. Convert every whitespace
 * in str to plain space. Return pointer to str. If !str or !*str or str only
 * contain whitespace, return NULL.
 */
char *strclean(char *str)
{
    char *p, *q;

    if (!str || !*str)
	return NULL;

    /* remove leading whitespace */
    p = str;
    while (*p && isspace((unsigned char) *p))	/* find the first non-blank */
	p++;
    if (!*p)			/* str was only blanks */
	return NULL;
    q = p;			/* save the start position of str */
    while ((*str++ = *p++))	/* copy from there */
	;

    /* remove trailing whitespace */
    p = q + strlen(q) - 1;	/* p points to the last character (before \0) */
    while (p > q && isspace((unsigned char) *p))
	p--;
    *(p + 1) = '\0';


    /* squeeze every sequence of whitespace to one whitespace */
    /* make every whitespace in the string be a normal space */

    str = p = q;		/* make str and p point to the current start of the string */

    while (*p) {
	if (isspace((unsigned char) *p)) {	/* if whitespace */
	    *str++ = ' ';	/* then make it a normal space */
	    while (isspace(*++p))	/* then skip until next non-ws */
		;
	    continue;		/* then start over again */
	}
	*str++ = *p++;		/* if not whitespace, then copy */
    }

    return *q ? q : NULL;
}

/*
 * Exit curses, print error message and call exit().
 */
void error_exit(char *fmt,...)
{
    va_list ap;

    CLEARLINE(MSG_LINE);
    refresh();
    endwin();
    if (fmt) {
	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
    }
    putchar('\n');
    exit(EXIT_FAILURE);
}

/*
 * If the first character of `s' is '~' or isn't '/', expand `s' and return
 * a pointer to `s'.
 * Paths longer than PATH_MAX will be silently truncated.
 * Return NULL on error.
 */
char *expand_path(char *s, size_t slen)
{
    char p[PATH_MAX + 1] = "";
    char *q = NULL;

    assert(s && *s);

    if (*s == '~') {
	if (*(s + 1) == '/')	/* s starts with ~/ */
	    snprintf(p, sizeof(p), "%s/%s", HomeDir, s + 2);	/* HomeDir must not end with slash */
	else if (*(s + 1) == '\0')	/* s is only a ~ */
	    snprintf(p, sizeof(p), "%s", HomeDir);
	else {			/* s is ~username */
	    struct passwd *pw;

	    q = strchr(s + 1, '/');
	    if (q)
		*q = 0;
	    if ((pw = getpwnam(s + 1)) && pw->pw_dir && pw->pw_dir[0]) {
		if (pw->pw_dir[strlen(pw->pw_dir) - 1] == '/')	/* is the last char a slash? */
		    pw->pw_dir[strlen(pw->pw_dir) - 1] = '\0';	/* if it is, then remove it */
		snprintf(p, sizeof(p), "%s/%s", pw->pw_dir, q ? q + 1 : "");
	    } else {		/* can't find home dir */
		if (q)		/* if we changed `s', */
		    *q = '/';	/* then undo it */
		return NULL;
	    }
	}
    } else if (s[0] != '/') {
      /* Append `s' to the current working directory and copy it into `s'. */
        strfcpy(p, s, sizeof(p)); /* save `s' */
	snprintf(s, slen, "%s%s%s",
		 cwd(),
		 IS_ROOT_PATH(cwd()) ? "" : "/",
		 p);
	STRIP_TRAILING_SLASHES_IF_NOT_ROOT_DIR(s);
	return s;
    } else {
        STRIP_TRAILING_SLASHES_IF_NOT_ROOT_DIR(s);
        return s;
    }

    strfcpy(s, p, slen);	/* replace the string with the expanded version. */

    STRIP_TRAILING_SLASHES_IF_NOT_ROOT_DIR(s);
    return s;
}

/*
 * Remove leading and trailing whitespace from str.
 * Return pointer to str if *str, return NULL if !*str.
 */

char *rm_leading_and_trailing_ws(char *str)
{
    char *p, *tmp;

    if (!str || !*str)
	return NULL;

    /* remove leading whitespace */
    p = str;
    while (whitespace(*p))	/* find the first non-blank */
	p++;

    if (!*p)
	return NULL;		/* str was only blanks */

    tmp = str;			/* save the address of the first non-blank */
    while ((*str++ = *p++))	/* copy from there */
	;
    str = tmp;			/* restore the address */
    /* remove trailing whitespace */

    p = str + strlen(str) - 1;
    /* p now points to the last character (before \0) */
    while (p > str && whitespace(*p))
	p--;
    *(p + 1) = '\0';

    return str;
}
